""" Module ncurses.form - a python binding to the ncurses form library.
## Copyright (c) 2000 by Harry Henry Gebel

This module is released under an MIT style license. See source for license
terms. """

## Permission is hereby granted, free of charge, to any person
## obtaining a copy of this software and associated documentation files
## (the "Software"), to deal in the Software without restriction,
## including without limitation the rights to use, copy, modify, merge,
## publish, distribute, sublicense, and/or sell copies of the Software,
## and to permit persons to whom the Software is furnished to do so,
## subject to the following conditions:

## The above copyright notice and this permission notice shall be
## included in all copies or substantial portions of the Software.

## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
## EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
## MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
## NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
## BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
## ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
## CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
## SOFTWARE.

from ncurses.panel import *
from ncurses import _form
from ncurses import _curses
from types import TupleType, ListType, StringType

# import constants
from _form import NO_JUSTIFICATION, JUSTIFY_LEFT, JUSTIFY_CENTER, \
     JUSTIFY_RIGHT, O_VISIBLE, O_ACTIVE, O_PUBLIC, O_EDIT, O_WRAP, O_BLANK, \
     O_AUTOSKIP, O_NULLOK, O_PASSOK, O_STATIC, O_NL_OVERLOAD, O_BS_OVERLOAD, \
     REQ_NEXT_PAGE, REQ_PREV_PAGE, REQ_FIRST_PAGE, REQ_LAST_PAGE, \
     REQ_NEXT_FIELD, REQ_PREV_FIELD, REQ_FIRST_FIELD, REQ_LAST_FIELD, \
     REQ_SNEXT_FIELD, REQ_SPREV_FIELD, REQ_SFIRST_FIELD, REQ_SLAST_FIELD, \
     REQ_LEFT_FIELD, REQ_RIGHT_FIELD, REQ_UP_FIELD, REQ_DOWN_FIELD, \
     REQ_NEXT_CHAR, REQ_PREV_CHAR, REQ_NEXT_LINE, REQ_PREV_LINE, \
     REQ_NEXT_WORD, REQ_PREV_WORD, REQ_BEG_FIELD, REQ_END_FIELD, \
     REQ_BEG_LINE, REQ_END_LINE, REQ_LEFT_CHAR, REQ_RIGHT_CHAR, REQ_UP_CHAR, \
     REQ_DOWN_CHAR, REQ_NEW_LINE, REQ_INS_CHAR, REQ_INS_LINE, REQ_DEL_CHAR, \
     REQ_DEL_PREV, REQ_DEL_LINE, REQ_DEL_WORD, REQ_CLR_EOL, REQ_CLR_EOF, \
     REQ_CLR_FIELD, REQ_OVL_MODE, REQ_INS_MODE, REQ_SCR_FLINE, REQ_SCR_BLINE, \
     REQ_SCR_FPAGE, REQ_SCR_BPAGE, REQ_SCR_FHPAGE, REQ_SCR_BHPAGE, \
     REQ_SCR_FCHAR, REQ_SCR_BCHAR, REQ_SCR_HFLINE, REQ_SCR_HBLINE, \
     REQ_SCR_HFHALF, REQ_SCR_HBHALF, REQ_VALIDATION, REQ_NEXT_CHOICE, \
     REQ_PREV_CHOICE, MIN_FORM_COMMAND, MAX_FORM_COMMAND

# import functions
from _form import form_request_name, form_request_by_name

# globals

_CURSES_MODE = 1
_PANEL_MODE = 2
_update_mode = _CURSES_MODE

def form_get_update_mode():
  global _update_mode
  if _update_mode == _PANEL_MODE:
    return 'panel'
  else:
    return 'curses'

def form_set_update_mode(new_mode):
  global _update_mode
  if new_mode != 'panel' and new_mode != 'curses':
    raise ValueError('Unknown update mode: ' + _update_mode)
  if new_mode == 'panel':
    _update_mode = _PANEL_MODE
  else:
    _update_mode = _CURSES_MODE
  
# wrappers

def doupdate():
  if _update_mode == _PANEL_MODE:
    update_panels()
  else:
    _curses.doupdate()

# utility functions
def _testSequence(seq):
  if type(seq) != TupleType and type(seq) != ListType:
    raise TypeError('Not a list or tuple: ' + str(seq))

# Class FIELD represents FIELD objects
class FIELD:
  def __init__(self, a, b, c, d = None, e = 0, f = 0, function = 'new_field'):
    if function == 'new_field':
      if d == None:
        raise TypeError('function "new_field" missing parameter(s)')
      self.FIELD = _form.new_field(a, b, c, d, e, f)
    elif function == 'dup_field':
      if d != None:
        raise TypeError('function "dup_field" requires three parameters')
      self.FIELD = _form.dup_field(a, b, c)
    elif function == 'link_field':
      if d != None:
        raise TypeError('function "link_field" requires three parameters')
      self.FIELD = _form.link_field(a, b, c)
    self.userdata = None

  def __makeArray(self, strings):
    """ Takes a list or tuple of strings and returns a null-terminated array
    of null-terminated character strings. Elements that are not strings will be
    converted to strings."""
    _testSequence(strings)
    # make array, including room for NULL pointer
    array = _curses.ptrcreate('int', 0, len(strings) + 1)
    # convert arguments to strings, create C strings and place in array
    for counter in xrange(len(strings)):
      if type(strings[counter]) != StringType: # make sure it's a string
        strings[counter] = str(strings[counter])
      _curses.ptrset(array, eval('0x' +
                                 _curses.ptrcreate('char',
                                                   strings[counter],
                                                   len(strings[counter]) +
                                                   2)[1:-7]), counter)
    array = _curses.ptrcast(array, 'char **') # cast to C string array
    return array

  def __freeArray(self):
    """ Frees any character string arrays created for validation. Does nothing
    if no character array exists. Also frees validation types """
    # detect if an validation type exists and free it if it does
    fld_type = _form.field_type(self.FIELD)
    if fld_type != None:
      _form.free_fieldtype(fld_type)
    fld_userptr = _form.field_userptr(self.FIELD)
    if fld_userptr != None:
      _form.free_validation_callback(fld_userptr)
    # detect if an array exists and free if it does
    if self.__dict__.has_key('array'):
      _curses.ptrfree(self.array)
      del self.array # remove referance to non-existant array

  def __del__(self):
    if _form:
      self.__freeArray() # make sure any validation arrays are released
      _form.free_field(self.FIELD)

  def dup_field(self, toprow, leftcol):
    return FIELD(self.FIELD, toprow, leftcol, function = 'dup_field')

  def link_field(self, toprow, leftcol):
    return FIELD(self.FIELD, toprow, leftcol, function = 'link_field')

  # return a tuple (rows, cols, frow, fcol, nrow, nbuf)
  def field_info(self):
    rows_p = _curses.ptrcreate('int')
    cols_p = _curses.ptrcreate('int')
    frow_p = _curses.ptrcreate('int')
    fcol_p = _curses.ptrcreate('int')
    nrow_p = _curses.ptrcreate('int')
    nbuf_p = _curses.ptrcreate('int')
    _form.field_info(self.FIELD, rows_p, cols_p, frow_p, fcol_p, nrow_p,
                     nbuf_p)
    rows = _curses.ptrvalue(rows_p)
    cols = _curses.ptrvalue(cols_p)
    frow = _curses.ptrvalue(frow_p)
    fcol = _curses.ptrvalue(fcol_p)
    nrow = _curses.ptrvalue(nrow_p)
    nbuf = _curses.ptrvalue(nbuf_p)
    _curses.ptrfree(rows_p)
    _curses.ptrfree(cols_p)
    _curses.ptrfree(frow_p)
    _curses.ptrfree(fcol_p)
    _curses.ptrfree(nrow_p)
    _curses.ptrfree(nbuf_p)
    return (rows, cols, frow, fcol, nrow, nbuf)

  # returns a tuple (rows, cols, max)
  def dynamic_field_info(self, a, b, c, d):
    rows_p = _curses.ptrcreate('int')
    cols_p = _curses.ptrcreate('int')
    max_p = _curses.ptrcreate('int')
    _form.field_info(self.FIELD, rows_p, cols_p, max_p)
    rows = _curses.ptrvalue(rows_p)
    cols = _curses.ptrvalue(cols_p)
    max = _curses.ptrvalue(max_p)
    _curses.ptrfree(rows_p)
    _curses.ptrfree(cols_p)
    _curses.ptrfree(max_p)
    return (rows, cols, max)

  def set_max_field(self, a): # checked
    return _form.set_max_field(self.FIELD, a)

  def move_field(self, a, b): # checked
    return _form.move_field(self.FIELD, a, b)

  def set_new_page(self, a): # checked
    return _form.set_new_page(self.FIELD, a)

  def set_field_just(self, a): # checked
    return _form.set_field_just(self.FIELD, a)

  def field_just(self): # checked
    return _form.field_just(self.FIELD)

  def set_field_fore(self, attr):
    return _form.set_field_fore(self.FIELD, chtype(attr))

  def set_field_back(self, attr):
    return _form.set_field_back(self.FIELD, chtype(attr))

  def set_field_pad(self, pad):
    if type(pad) == type(''):
      return _form.set_field_pad(self.FIELD, ord(pad))
    else:
      return _form.set_field_pad(self.FIELD, pad)

  def field_pad(self):
    pad = _form.field_pad(self.FIELD)
    if pad < 256: # convert to a single character string if possible
      return chr(pad)
    else:
      return pad

  # arguments are backwards
  def set_field_buffer(self, value, buf = 0):
    return _form.set_field_buffer(self.FIELD, buf, value)

  def set_field_status(self, a): # checked
    return _form.set_field_status(self.FIELD, a)

  def set_field_opts(self, a): # checked
    return _form.set_field_opts(self.FIELD, a)

  def field_opts_on(self, a): # checked
    return _form.field_opts_on(self.FIELD, a)

  def field_opts_off(self, a): # checked
    return _form.field_opts_off(self.FIELD, a)

  def field_fore(self): # checked
    return _form.field_fore(self.FIELD)

  def field_back(self): #checked
    return _form.field_back(self.FIELD)

  def new_page(self): # checked
    return _form.new_page(self.FIELD)

  def field_status(self): # checked
    return _form.field_status(self.FIELD)

  def field_buffer(self, buf = 0): 
    return _form.field_buffer(self.FIELD, buf)

  def field_opts(self): # checked
    return _form.field_opts(self.FIELD)

  def field_index(self): # checked
    return _form.field_index(self.FIELD)

  def TYPE_ALPHA(self, min):
    """ Sets field validation to TYPE_ALPHA """
    self.__freeArray()
    return _form.field_TYPE_ALPHA(self.FIELD, min)

  def TYPE_ALNUM(self, min):
    """ Sets field validation to TYPE_ALNUM """
    self.__freeArray()
    return _form.field_TYPE_ALNUM(self.FIELD, min)

  def TYPE_ENUM(self, list, case = FALSE, match = FALSE):
    """ Sets field validation to TYPE_ENUM """
    self.__freeArray() # get rid of any old arrays before making a new one
    self.array = self.__makeArray(list) # make new array
    return _form.field_TYPE_ENUM(self.FIELD, self.array, case, match)

  def TYPE_INTEGER(self, precision, min, max):
    """ Sets field validation to TYPE_INTEGER """
    self.__freeArray()
    return _form.field_TYPE_INTEGER(self.FIELD, precision, min, max)

  def TYPE_NUMERIC(self, precision, min, max):
    """ Sets field validation to TYPE_NUMERIC """
    self.__freeArray()
    return _form.field_TYPE_NUMERIC(self.FIELD, precision, min, max)

  def TYPE_REGEXP(self, regexp):
    self.__freeArrary()
    return _form.field_TYPE_REGEXP(self.FIELD, regexp)

  def TYPE_IPV4(self):
    self.__freeArray()
    return _form.field_TYPE_IPV4(self.FIELD)

  def set_validater(self,fld_chk,chr_chk=None):
    """ register python callback function """
    return _form.set_callback(self.FIELD,self,fld_chk,chr_chk)

# Class LABEL represents FIELD objects used as labels. IMPORTANT note the order
# or arguments
class LABEL(FIELD):
  def __init__(self, toprow, leftcol, label, width = 0,
               function = 'new_field', original = None):
    if function == 'new_field':
      if width == 0: # calculate width
        if len(label) == 0:
          raise ncursesError('Label must contain a value if width is zero')
        width = len(label)
      FIELD.__init__(self, 1, width, toprow, leftcol) # make field
      self.field_opts_off(O_ACTIVE) # turn into a label
      self.set_field_buffer(label)
      self.set_field_fore(A_BOLD)
    elif function == 'dup_field':
      FIELD.__init__(self, original, toprow, leftcol, function = 'dup_field')
    elif function == 'link_field':
      FIELD.__init__(self, original, toprow, leftcol, function = 'link_field')
    else:
      raise ncursesError('Unknown initialization function')

  def dup_field(self, toprow, leftcol):
    return LABEL(toprow, leftcol, None, None, None, 'dup_field', self.FIELD)

  def link_field(self, toprow, leftcol):
    return LANEL(toprow, leftcol, None, None, None, 'link_field', self.FIELD)

# Class FORM represents FORM objects
class FORM:
  def __init__(self, fields):
    _testSequence(fields)
    # generate C array of pointers to ITEMS and create menu
    self.__fields = self.__makeArray(fields)
    self.__fieldStorage = tuple(fields) # protects fields from deallocation
    self.FORM = _form.new_form(self.__fields)

  def __del__(self):
    if _form.free_form(self.FORM) != E_OK: 
      self.unpost_form() # unpost if already posted
      _form.free_form(self.FORM)
    _curses.ptrfree(self.__fields) # deallocate item array

  # Makes a C array of pointers to FIELDs
  def __makeArray(self, fields):
    array = _curses.ptrcreate('int', 0, len(fields) + 1) # allocate array
    for field in xrange(len(fields)):
      _curses.ptrset(array, eval('0x' + fields[field].FIELD[1:-8]), field)
    return _curses.ptrcast(array, 'FIELD **')

  def current_field(self, fields = None):
    """ takes a list of FIELDs and returns the current field """
    if not fields:
      fields = self.__fieldStorage
    current = _form.current_field(self.FORM)
    for field in fields:
      if field.FIELD == current:
        return field
    raise ncursesError('Unable to find current field in list')

  def set_form_fields(self, fields):
    """ replaces old field list with a new one """
    newfields = self.__makeArray(fields)
    if _form.set_form_fields(self.FORM, newfields) == E_OK:
      _curses.ptrfree(self.__fields)
      self.__fields = newfields
      self.__fieldStorage = tuple(fields)
    else:
      raise ncursesError('Unable to set fields')

  def field_count(self): # checked
    return _form.field_count(self.FORM)

  def set_form_win(self, win):
    return _form.set_form_win(self.FORM, win.WINDOW)

  def set_form_sub(self, sub):
    return _form.set_form_sub(self.FORM, sub.WINDOW)

  def set_current_field(self, field):
    return _form.set_current_field(self.FORM, field.FIELD)

  def set_form_page(self, a): # checked
    return _form.set_form_page(self.FORM, a)

  def form_page(self): # checked
    return _form.form_page(self.FORM)

  # returns a tuple (rows, columns)
  def scale_form(self):
    rows_p = _curses.ptrcreate('int')
    columns_p = _curses.ptrcreate('int')
    _form.scale_form(self.FORM, rows_p, columns_p)
    rows = _curses.ptrvalue(rows_p)
    columns = _curses.ptrvalue(columns_p)
    _curses.ptrfree(rows_p)
    _curses.ptrfree(columns_p)
    return (rows, columns)

  def post_form(self): # checked
    return _form.post_form(self.FORM)

  def unpost_form(self): # checked
    return _form.unpost_form(self.FORM)

  def pos_form_cursor(self): # checked
    return _form.pos_form_cursor(self.FORM)

  def form_driver(self, a): # checked
    return _form.form_driver(self.FORM, a)

  def set_form_opts(self, a): # checked
    return _form.set_form_opts(self.FORM, a)

  def form_opts_on(self, a): # checked
    return _form.form_opts_on(self.FORM, a)

  def form_opts_off(self, a): # checked
    return _form.form_opts_off(self.FORM, a)

  def form_opts(self): # checked
    return _form.form_opts(self.FORM)

  def data_ahead(self): # checked
    return _form.data_ahead(self.FORM)

  def data_behind(self): # checked
    return _form.data_behind(self.FORM)
