## Copyright (c) 2000 by Harry Henry Gebel

## Permission is hereby granted, free of charge, to any person
## obtaining a copy of this software and associated documentation files
## (the "Software"), to deal in the Software without restriction,
## including without limitation the rights to use, copy, modify, merge,
## publish, distribute, sublicense, and/or sell copies of the Software,
## and to permit persons to whom the Software is furnished to do so,
## subject to the following conditions:

## The above copyright notice and this permission notice shall be
## included in all copies or substantial portions of the Software.

## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
## EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
## MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
## NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
## BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
## ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
## CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
## SOFTWARE.

from ncurses.panel import *
from ncurses import _menu
from ncurses import _curses
from types import ListType, TupleType

# import constants
from _menu import O_ONEVALUE, O_SHOWDESC, O_ROWMAJOR, O_IGNORECASE, \
     O_SHOWMATCH, O_NONCYCLIC, O_SELECTABLE, REQ_LEFT_ITEM, REQ_RIGHT_ITEM, \
     REQ_UP_ITEM, REQ_DOWN_ITEM, REQ_SCR_ULINE, REQ_SCR_DLINE, REQ_SCR_DPAGE, \
     REQ_SCR_UPAGE, REQ_FIRST_ITEM, REQ_LAST_ITEM, REQ_NEXT_ITEM, \
     REQ_PREV_ITEM, REQ_TOGGLE_ITEM, REQ_CLEAR_PATTERN, REQ_BACK_PATTERN, \
     REQ_NEXT_MATCH, REQ_PREV_MATCH, MIN_MENU_COMMAND, MAX_MENU_COMMAND

# globals

_CURSES_MODE = 1
_PANEL_MODE = 2
_update_mode = _CURSES_MODE

def menu_get_update_mode():
  global _update_mode
  if _update_mode == _PANEL_MODE:
    return 'panel'
  else:
    return 'curses'

def menu_set_update_mode(new_mode):
  global _update_mode
  if new_mode != 'panel' and new_mode != 'curses':
    raise ValueError('Unknown update mode: ' + _update_mode)
  if new_mode == 'panel':
    _update_mode = _PANEL_MODE
  else:
    _update_mode = _CURSES_MODE
  
# wrapperss

def doupdate():
  if _update_mode == _PANEL_MODE:
    update_panels()
  else:
    _curses.doupdate()

# utility functions
def _testSequence(seq):
  if type(seq) != TupleType and type(seq) != ListType:
    raise TypeError('Not a list or tuple: ' + str(seq))


# class ITEM - handles ITEM objects
class ITEM:
  def __init__(self, name, description):
    if type(name) != type('') or type(description) != type(''):
      raise TypeError('ITEM must be initialized with two strings')
    # ITEM does not store strings, so we will
    self.name_p = _curses.ptrcreate('char', name, len(name) + 2)
    self.name = name
    self.description_p = _curses.ptrcreate('char', description,
                                           len(description) + 2)
    self.description = description
    # a little trick to make SWIG pass the pointer instead of a copy of the
    # string representing the pointer
    self.ITEM = _menu.void_new_item(self.name_p, self.description_p)
    self.user_data = None

  def __del__(self):
    if _menu:
      _menu.free_item(self.ITEM)
      _curses.ptrfree(self.name_p)
      _curses.ptrfree(self.description_p)

  def item_opts(self):
    return _menu.item_opts(self.ITEM)

  def item_index(self):
    return _menu.item_index(self.ITEM)

  def item_opts_off(self, a):
    return _menu.item_opts_off(self.ITEM, a)

  def item_opts_on(self, a):
    return _menu.item_opts_on(self.ITEM, a)


  def set_item_opts(self, a):
    return _menu.set_item_opts(self.ITEM, a)

  def set_item_value(self, a):
    return _menu.set_item_value(self.ITEM, a)

  def item_value(self):
    return _menu.item_value(self.ITEM)

  def item_visible(self):
    return _menu.item_visible(self.ITEM)

# class MENU implements MENU objects
class MENU:
  def __init__(self, items):
    _testSequence(items)
    # generate C array of pointers to ITEMS and create menu
    self.__items = self.__makeArray(items)
    self.__itemStorage = tuple(items) # protect items from deallocation
    self.MENU = _menu.new_menu(self.__items)
    self.user_data = None

  def __del__(self):
    if _menu:
      if _menu.free_menu(self.MENU) != E_OK:
        self.unpost_menu()
        _menu.free_menu(self.MENU)
      _curses.ptrfree(self.__items) # deallocate item array

  # Makes a C array of pointers to ITEMs
  def __makeArray(self, items):
    array = _curses.ptrcreate('int', 0, len(items) + 1) # allocate array
    for item in xrange(len(items)):
      _curses.ptrset(array, eval('0x' + items[item].ITEM[1:-7]), item)
    return _curses.ptrcast(array, 'ITEM **')

  def set_menu_items(self, items):
    """ replaces old menu item list with a new one """
    olditems = self.__items
    self.__items = self.makeArray(items)
    if _menu.set_menu_items(self.__items) == E_OK:
      _curses.ptrfree(olditems) # deallocate old array
      self.__itemStorage = tuple(items)
    else:
      _curses.ptrfree(self.__items) # deallocate new array
      self.__items = olditems
      raise ncursesError('Failed to set new items')

  def set_current_item(self, item):
    return _menu.set_current_item(self.MENU, item.ITEM)

  def current_item(self, items = None):
    """ Return the current item """
    if not items:
      items = self.__itemStorage
    current = _menu.current_item(self.MENU)
    for item in items:
      if item.ITEM == current:
        return item
    raise ncursesError('Current item not found')

  def menu_opts(self):
    return _menu.menu_opts(self.MENU)

  def menu_mark(self):
    return _menu.menu_mark(self.MENU)

  def menu_pattern(self):
    return _menu.menu_pattern(self.MENU)

  def menu_back(self):
    return _menu.menu_back(self.MENU)

  def menu_fore(self):
    return _menu.menu_fore(self.MENU)

  def menu_grey(self):
    return _menu.menu_grey(self.MENU)

  def item_count(self):
    return _menu.item_count(self.MENU)

  def menu_driver(self, a):
    return _menu.menu_driver(self.MENU, a)

  def menu_opts_off(self, a):
    return _menu.menu_opts_off(self.MENU, a)

  def menu_opts_on(self, a):
    return _menu.menu_opts_on(self.MENU, a)

  def menu_pad(self):
    return _menu.menu_pad(self.MENU)

  def pos_menu_cursor(self):
    return _menu.pos_menu_cursor(self.MENU)

  def post_menu(self):
    return _menu.post_menu(self.MENU)

  # returns a tuple (rows, columns)
  def scale_menu(self):
    rows_p = _curses.ptrcreate('int') # make pointers
    columns_p = _curses.ptrcreate('int')
    _menu.scale_menu(self.MENU, rows_p, columns_p)
    rows = _curses.ptrvalue(rows_p)
    columns = _curses.ptrvalue(columns_p)
    _curses.ptrfree(rows_p) # destroy pointers
    _curses.ptrfree(columns_p)
    return (rows, columns)

  def set_menu_back(self, attr):
    return _menu.set_menu_back(self.MENU, chtype(attr))

  def set_menu_fore(self, attr):
    return _menu.set_menu_fore(self.MENU, chtype(attr))

  def set_menu_format(self, a, b):
    return _menu.set_menu_format(self.MENU, a, b)

  def set_menu_grey(self, attr):
    return _menu.set_menu_grey(self.MENU, chtype(attr))

  def set_menu_mark(self, mark):
    if mark == '' or mark == None: # set mark to '' or None to abolish mark
      return _menu.set_menu_mark(self.MENU, 'NULL')
    else:
      return _menu.set_menu_mark(self.MENU, mark)

  def set_menu_opts(self, a):
    return _menu.set_menu_opts(self.MENU, a)

  def set_menu_pad(self, attr):
    return _menu.set_menu_pad(self.MENU, chtype(attr))

  def set_menu_pattern(self, a):
    return _menu.set_menu_pattern(self.MENU, a)

  def set_menu_sub(self, window):
    if type(window) == type(''): # used to pass pointers as string, for NULL
      return _menu.set_menu_sub(self.MENU, window)
    else:
      return _menu.set_menu_sub(self.MENU, window.WINDOW)

  def set_menu_win(self, window):
    if type(window) == type(''): # used to pass pointers as string, for NULL
      return _menu.set_menu_win(self.MENU, window)
    else:
      return _menu.set_menu_win(self.MENU, window.WINDOW)

  def set_top_row(self, a):
    return _menu.set_top_row(self.MENU, a)

  def top_row(self):
    return _menu.top_row(self.MENU)

  def unpost_menu(self):
    return _menu.unpost_menu(self.MENU)

  def set_menu_spacing(self, a, b, c):
    return _menu.set_menu_spacing(self.MENU, a, b, c)

  # returns a tuple (spc_description, spc_row, spc_columns)
  def menu_spacing(self):
    desc_p = _curses.ptrcreate('int')
    row_p = _curses.ptrcreate('int')
    col_p = _curses.ptrcreate('int')
    _menu.menu_spacing(self.MENU, desc_p, row_p, col_p)
    desc = _curses.ptrvalue(desc_p)
    row = _curses.ptrvalue(row_p)
    col = _curses.ptrvalue(col_p)
    _curses.ptrfree(desc_p)
    _curses.ptrfree(row_p)
    _curses.ptrfree(col_p)
    return (desc, row, col)

  # return a tuple (rows, columns)
  def menu_format(self):
    rows_p = _curses.ptrcreate('int') # make pointers
    columns_p = _curses.ptrcreate('int')
    _menu.menu_format(self.MENU, rows_p, columns_p)
    rows = _curses.ptrvalue(rows_p)
    columns = _curses.ptrvalue(columns_p)
    _curses.ptrfree(rows_p) # destroy pointers
    _curses.ptrfree(columns_p)
    return (rows, columns)
